/*
 * Copyright@ Samsung Electronics Co. LTD
 *
 * This software is proprietary of Samsung Electronics.
 * No part of this software, either material or conceptual may be copied or distributed, transmitted,
 * transcribed, stored in a retrieval system or translated into any human or computer language in any form by any means,
 * electronic, mechanical, manual or otherwise, or disclosed
 * to third parties without the express written permission of Samsung Electronics.
 *
 * Alternatively, this program is free software in case of open source project
 * you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "bl2_config.h"
#include "board_config.h"
#include "bl2_common.h"
#include "ect_parser.h"
#include "smc_aarch64.h"

#define ALIGNMENT_SIZE	 4
#define ALLOC_ALIGNMENT_SIZE	8

/* Variable */

static char g_signature[] = "PARA";

/* API for internal */

static void* g_base_alloc;

static int strcmp_internal(const char * cs,const char * ct)
{
	register signed char __res;

	while (1) {
		if ((__res = *cs - *ct++) != 0 || !*cs++)
			break;
	}

	return __res;
}

static int strncmp_internal(const char * cs,const char * ct,unsigned int count)
{
	register signed char __res = 0;

	while (count) {
		if ((__res = *cs - *ct++) != 0 || !*cs++)
			break;
		count--;
	}

	return __res;
}

static int memcmp_internal(const void * cs,const void * ct,unsigned int count)
{
	const unsigned char *su1, *su2;
	int res = 0;

	for( su1 = cs, su2 = ct; 0 < count; ++su1, ++su2, count--)
		if ((res = *su1 - *su2) != 0)
			break;
	return res;
}

static void align_address(void)
{
	if ((unsigned long)g_base_alloc % ALLOC_ALIGNMENT_SIZE != 0)
		g_base_alloc += ALLOC_ALIGNMENT_SIZE - ((unsigned long)g_base_alloc % ALLOC_ALIGNMENT_SIZE);
}

static void* kzalloc(unsigned int size)
{
	void *ret = g_base_alloc;

	g_base_alloc += size;
	align_address();

	return ret;
}

static void ect_parse_integer(void **address, void *value)
{
	*((unsigned int *)value) = __raw_readl(*address);
	*address += sizeof(unsigned int);
}

static int ect_parse_string(void **address, char **value, unsigned int *length)
{
	ect_parse_integer(address, length);
	(*length)++;

	*value = *address;

	if (*length % ALIGNMENT_SIZE != 0)
		*address += *length + ALIGNMENT_SIZE - (*length % ALIGNMENT_SIZE);
	else
		*address += *length;

	return 0;
}

static int ect_parse_dvfs_domain(int parser_version, void *address, struct ect_dvfs_domain *domain)
{
	int i;
	char *clock_name;
	unsigned int length;

	ect_parse_integer(&address, &domain->max_frequency);
	ect_parse_integer(&address, &domain->min_frequency);

	if (parser_version >= 2) {
		ect_parse_integer(&address, &domain->boot_level_idx);
		ect_parse_integer(&address, &domain->resume_level_idx);
	} else {
		domain->boot_level_idx = -1;
		domain->resume_level_idx = -1;
	}

	ect_parse_integer(&address, &domain->num_of_clock);
	ect_parse_integer(&address, &domain->num_of_level);

	domain->list_clock = kzalloc(sizeof(char *) * domain->num_of_clock);
	if (domain->list_clock == NULL)
		return -2;

	for (i = 0; i < domain->num_of_clock; ++i) {
		if (ect_parse_string(&address, &clock_name, &length))
			return -1;

		domain->list_clock[i] = clock_name;
	}

	domain->list_level = address;
	address += sizeof(struct ect_dvfs_level) * domain->num_of_level;

	domain->list_dvfs_value = address;

	return 0;
}

static int ect_parse_dvfs_header(void *address, struct ect_info *info)
{
	int i;
	char *domain_name;
	unsigned int length, offset;
	struct ect_dvfs_header *ect_dvfs_header;
	struct ect_dvfs_domain *ect_dvfs_domain;
	void *address_dvfs_header = address;

	if (address == NULL)
		return -1;

	ect_dvfs_header = kzalloc(sizeof(struct ect_dvfs_header));
	if (ect_dvfs_header == NULL)
		return -2;

	ect_parse_integer(&address, &ect_dvfs_header->parser_version);
	ect_parse_integer(&address, &ect_dvfs_header->version);
	ect_parse_integer(&address, &ect_dvfs_header->num_of_domain);

	ect_dvfs_header->domain_list = kzalloc(sizeof(struct ect_dvfs_domain) * ect_dvfs_header->num_of_domain);
	if (ect_dvfs_header->domain_list == NULL)
		return -1;

	for (i = 0; i < ect_dvfs_header->num_of_domain; ++i) {
		if (ect_parse_string(&address, &domain_name, &length))
			return -1;

		ect_parse_integer(&address, &offset);

		ect_dvfs_domain = &ect_dvfs_header->domain_list[i];
		ect_dvfs_domain->domain_name = domain_name;
		ect_dvfs_domain->domain_offset = offset;
	}

	for (i = 0; i < ect_dvfs_header->num_of_domain; ++i) {
		ect_dvfs_domain = &ect_dvfs_header->domain_list[i];

		if (ect_parse_dvfs_domain(ect_dvfs_header->parser_version,
					address_dvfs_header + ect_dvfs_domain->domain_offset,
					ect_dvfs_domain))
			return -1;
	}

	info->block_handle = ect_dvfs_header;

	return 0;
}

static int ect_parse_pll(int parser_version, void *address, struct ect_pll *ect_pll)
{
	ect_parse_integer(&address, &ect_pll->type_pll);
	ect_parse_integer(&address, &ect_pll->num_of_frequency);

	ect_pll->frequency_list = address;

	return 0;
}

static int ect_parse_voltage_table(int parser_version, void **address, struct ect_voltage_domain *domain, struct ect_voltage_table *table)
{
	int num_of_data = domain->num_of_group * domain->num_of_level;

	ect_parse_integer(address, &table->table_version);

	if (parser_version >= 2) {
		ect_parse_integer(address, &table->boot_level_idx);
		ect_parse_integer(address, &table->resume_level_idx);

		table->level_en = *address;
		*address += sizeof(int) * domain->num_of_level;
	} else {
		table->boot_level_idx = -1;
		table->resume_level_idx = -1;

		table->level_en = NULL;
	}

	if (parser_version >= 3) {
		table->voltages = NULL;

		table->voltages_step = *address;
		*address += sizeof(unsigned char) * num_of_data;
		table->volt_step = PMIC_VOLTAGE_STEP;
	} else {
		table->voltages = *address;
		*address += sizeof(int) * num_of_data;

		table->voltages_step = NULL;
		table->volt_step = 0;
	}

	return 0;
}

static int ect_parse_voltage_domain(int parser_version, void *address, struct ect_voltage_domain *domain)
{
	int i;

	ect_parse_integer(&address, &domain->num_of_group);
	ect_parse_integer(&address, &domain->num_of_level);
	ect_parse_integer(&address, &domain->num_of_table);

	domain->level_list = address;
	address += sizeof(int) * domain->num_of_level;

	domain->table_list = kzalloc(sizeof(struct ect_voltage_table) * domain->num_of_table);
	if (domain->table_list == NULL)
		return -2;

	for (i = 0; i < domain->num_of_table; ++i) {
		if (ect_parse_voltage_table(parser_version,
					&address,
					domain,
					&domain->table_list[i]))
			return -1;
	}

	return 0;
}

static int ect_parse_voltage_header(void *address, struct ect_info *info)
{
	int i;
	char *domain_name;
	unsigned int length, offset;
	struct ect_voltage_header *ect_voltage_header;
	struct ect_voltage_domain *ect_voltage_domain;
	void *address_voltage_header = address;

	if (address == NULL)
		return -1;

	ect_voltage_header = kzalloc(sizeof(struct ect_voltage_header));
	if (ect_voltage_header == NULL)
		return -1;

	ect_parse_integer(&address, &ect_voltage_header->parser_version);
	ect_parse_integer(&address, &ect_voltage_header->version);
	ect_parse_integer(&address, &ect_voltage_header->num_of_domain);

	ect_voltage_header->domain_list = kzalloc(sizeof(struct ect_voltage_domain) * ect_voltage_header->num_of_domain);
	if (ect_voltage_header->domain_list == NULL)
		return -2;

	for (i = 0; i < ect_voltage_header->num_of_domain; ++i) {
		if (ect_parse_string(&address, &domain_name, &length))
			return -1;

		ect_parse_integer(&address, &offset);

		ect_voltage_domain = &ect_voltage_header->domain_list[i];
		ect_voltage_domain->domain_name = domain_name;
		ect_voltage_domain->domain_offset = offset;
	}

	for (i = 0; i < ect_voltage_header->num_of_domain; ++i) {
		ect_voltage_domain = &ect_voltage_header->domain_list[i];

		if (ect_parse_voltage_domain(ect_voltage_header->parser_version,
					address_voltage_header + ect_voltage_domain->domain_offset,
					ect_voltage_domain))
			return -1;
	}

	info->block_handle = ect_voltage_header;

	return 0;
}

static int ect_parse_rcc_table(int parser_version, void **address, struct ect_rcc_domain *domain, struct ect_rcc_table *table)
{
	int num_of_data = domain->num_of_group * domain->num_of_level;

	ect_parse_integer(address, &table->table_version);

	if (parser_version >= 2) {
		table->rcc_compact = *address;
		*address += sizeof(unsigned char) * num_of_data;
	} else {
		table->rcc = *address;
		*address += sizeof(int) * num_of_data;
	}

	return 0;
}

static int ect_parse_rcc_domain(int parser_version, void *address, struct ect_rcc_domain *domain)
{
	int i;

	ect_parse_integer(&address, &domain->num_of_group);
	ect_parse_integer(&address, &domain->num_of_level);
	ect_parse_integer(&address, &domain->num_of_table);

	domain->level_list = address;
	address += sizeof(int) * domain->num_of_level;

	domain->table_list = kzalloc(sizeof(struct ect_rcc_table) * domain->num_of_table);
	if (domain->table_list == NULL)
		return -2;

	for (i = 0; i < domain->num_of_table; ++i) {
		if (ect_parse_rcc_table(parser_version,
					&address,
					domain,
					&domain->table_list[i]))
			return -1;
	}

	return 0;
}

static int ect_parse_ap_thermal_function(int parser_version, void *address, struct ect_ap_thermal_function *function)
{
	int i;
	struct ect_ap_thermal_range *range;

	ect_parse_integer(&address, &function->num_of_range);

	function->range_list = kzalloc(sizeof(struct ect_ap_thermal_range) * function->num_of_range);

	for (i = 0; i < function->num_of_range; ++i) {
		range = &function->range_list[i];

		ect_parse_integer(&address, &range->lower_bound_temperature);
		ect_parse_integer(&address, &range->upper_bound_temperature);
		ect_parse_integer(&address, &range->max_frequency);
		ect_parse_integer(&address, &range->sw_trip);
		ect_parse_integer(&address, &range->flag);
	}

	return 0;
}

static int ect_parse_margin_domain(int parser_version, void *address, struct ect_margin_domain *domain)
{
	ect_parse_integer(&address, &domain->num_of_group);
	ect_parse_integer(&address, &domain->num_of_level);

	if (parser_version >= 2) {
		domain->offset = NULL;
		domain->offset_compact = address;
		domain->volt_step = PMIC_VOLTAGE_STEP;
	} else {
		domain->offset = address;
		domain->offset_compact = NULL;
	}

	return 0;
}

static int ect_parse_margin_header(void *address, struct ect_info *info)
{
	int i;
	char *domain_name;
	unsigned int length, offset;
	struct ect_margin_header *ect_margin_header;
	struct ect_margin_domain *ect_margin_domain;
	void *address_margin_header = address;

	if (address == NULL)
		return -1;

	ect_margin_header = kzalloc(sizeof(struct ect_margin_header));
	if (ect_margin_header == NULL)
		return -1;

	ect_parse_integer(&address, &ect_margin_header->parser_version);
	ect_parse_integer(&address, &ect_margin_header->version);
	ect_parse_integer(&address, &ect_margin_header->num_of_domain);

	ect_margin_header->domain_list = kzalloc(sizeof(struct ect_margin_domain) * ect_margin_header->num_of_domain);
	if (ect_margin_header->domain_list == NULL)
		return -2;
	for (i = 0; i < ect_margin_header->num_of_domain; ++i) {
		if (ect_parse_string(&address, &domain_name, &length))
			return -1;

		ect_parse_integer(&address, &offset);

		ect_margin_domain = &ect_margin_header->domain_list[i];
		ect_margin_domain->domain_name = domain_name;
		ect_margin_domain->domain_offset = offset;
	}

	for (i = 0; i < ect_margin_header->num_of_domain; ++i) {
		ect_margin_domain = &ect_margin_header->domain_list[i];

		if (ect_parse_margin_domain(ect_margin_header->parser_version,
					address_margin_header + ect_margin_domain->domain_offset,
					ect_margin_domain))
			return -1;
	}

	info->block_handle = ect_margin_header;

	return 0;
}

static int ect_parse_timing_param_size(int parser_version, void *address, struct ect_timing_param_size *size)
{
	ect_parse_integer(&address, &size->num_of_timing_param);
	ect_parse_integer(&address, &size->num_of_level);

	size->timing_parameter = address;

	return 0;
}

static int ect_parse_minlock_domain(int parser_version, void *address, struct ect_minlock_domain *domain)
{
	ect_parse_integer(&address, &domain->num_of_level);

	domain->level = address;

	return 0;
}

static int ect_parse_gen_param_size(int parser_version, void *address, struct ect_gen_param_table *size)
{
	ect_parse_integer(&address, &size->num_of_col);
	ect_parse_integer(&address, &size->num_of_row);

	size->parameter = address;

	return 0;
}

static int ect_parse_gen_param_header(void *address, struct ect_info *info)
{
	int i;
	char *table_name;
	unsigned int length, offset;
	struct ect_gen_param_header *ect_gen_param_header;
	struct ect_gen_param_table *ect_gen_param_table;
	void *address_param_header = address;

	if (address == NULL)
		return -1;

	ect_gen_param_header = kzalloc(sizeof(struct ect_gen_param_header));
	if (ect_gen_param_header == NULL)
		return -2;

	ect_parse_integer(&address, &ect_gen_param_header->parser_version);
	ect_parse_integer(&address, &ect_gen_param_header->version);
	ect_parse_integer(&address, &ect_gen_param_header->num_of_table);

	ect_gen_param_header->table_list = kzalloc(sizeof(struct ect_gen_param_table) * ect_gen_param_header->num_of_table);
	if (ect_gen_param_header->table_list == NULL)
		return -2;

	for (i = 0; i < ect_gen_param_header->num_of_table; ++i) {
		if (ect_parse_string(&address, &table_name, &length))
			return -1;

		ect_parse_integer(&address, &offset);

		ect_gen_param_table = &ect_gen_param_header->table_list[i];
		ect_gen_param_table->table_name = table_name;
		ect_gen_param_table->offset = offset;
	}

	for (i = 0; i < ect_gen_param_header->num_of_table; ++i) {
		ect_gen_param_table = &ect_gen_param_header->table_list[i];

		if (ect_parse_gen_param_size(ect_gen_param_header->parser_version,
					address_param_header + ect_gen_param_table->offset,
					ect_gen_param_table))
			return -1;
	}

	info->block_handle = ect_gen_param_header;

	return 0;
}


struct ect_header *g_ect_header;

static struct ect_info ect_list[] = {
	{
		.block_name = BLOCK_ASV,
		.block_name_length = sizeof(BLOCK_ASV) - 1,
		.parser = ect_parse_voltage_header,
		.block_handle = NULL,
	}, {
		.block_name = BLOCK_DVFS,
		.block_name_length = sizeof(BLOCK_DVFS) - 1,
		.parser = ect_parse_dvfs_header,
		.block_handle = NULL,
	}, {
		.block_name = BLOCK_MARGIN,
		.block_name_length = sizeof(BLOCK_MARGIN) - 1,
		.parser = ect_parse_margin_header,
		.block_handle = NULL,
	}
};

static void ect_set_bootsafe_status(int flag)
{
	writel(flag, PMU_SYSIP_DAT3);
}

static int ect_get_bootsafe_status(void)
{
	return readl(PMU_SYSIP_DAT3);
}

/* API for external */

static void *g_address;
static void *g_size;

void ect_init(void)
{
	unsigned long long ret = 0;
	int bootsafe_status;

	if (!((find_second_boot() & 0xF) == 1)) {
		ect_init_internal((void *)(CONFIG_SPL_TEXT_BASE + PART_SIZE_BL2_CODE + PART_SIZE_APM), (void *)PART_SIZE_ECT);
		ect_parse_binary_header();
		return ;
	}

	bootsafe_status = ect_get_bootsafe_status();

	if (bootsafe_status == BOOTSAFE_CLEAR) {
		load_ect_head();

		ret = ect_check_valid_header((void *)(CONFIG_SPL_TEXT_BASE + PART_SIZE_BL2_CODE + PART_SIZE_APM),
						(void *)CONFIG_PHY_IRAM_NS_BASE);
		if (!ret) {
			ect_set_bootsafe_status(BOOTSAFE_SET);
			load_ect_all();
		} else {
			skip_ect();
		}
	} else {
		load_ect_head();
		skip_ect();
	}

	ect_init_internal((void *)(CONFIG_SPL_TEXT_BASE + PART_SIZE_BL2_CODE + PART_SIZE_APM), (void *)PART_SIZE_ECT);
	ect_parse_binary_header();
}

void ect_init_internal(void *address, void *size)
{
	g_address = address;
	g_size = size;

	g_base_alloc = g_address;
}

void *ect_get_block(char *block_name)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(ect_list); ++i) {
		if (strcmp_internal(block_name, ect_list[i].block_name) == 0)
			return ect_list[i].block_handle;
	}

	return NULL;
}

struct ect_dvfs_domain *ect_dvfs_get_domain(void *block, char *domain_name)
{
	int i;
	struct ect_dvfs_header *header;
	struct ect_dvfs_domain *domain;

	if (block == NULL ||
		domain_name == NULL)
		return NULL;

	header = (struct ect_dvfs_header *)block;

	for (i = 0; i < header->num_of_domain; ++i) {
		domain = &header->domain_list[i];

		if (strcmp_internal(domain_name, domain->domain_name) == 0)
			return domain;
	}

	return NULL;
}

struct ect_voltage_domain *ect_asv_get_domain(void *block, char *domain_name)
{
	int i;
	struct ect_voltage_header *header;
	struct ect_voltage_domain *domain;

	if (block == NULL ||
		domain_name == NULL)
		return NULL;

	header = (struct ect_voltage_header *)block;

	for (i = 0; i < header->num_of_domain; ++i) {
		domain = &header->domain_list[i];

		if (strcmp_internal(domain_name, domain->domain_name) == 0)
			return domain;
	}

	return NULL;
}

struct ect_margin_domain *ect_margin_get_domain(void *block, char *domain_name)
{
	int i;
	struct ect_margin_header *header;
	struct ect_margin_domain *domain;

	if (block == NULL ||
		domain_name == NULL)
		return NULL;

	header = (struct ect_margin_header *)block;

	for (i = 0; i < header->num_of_domain; ++i) {
		domain = &header->domain_list[i];

		if (strcmp_internal(domain_name, domain->domain_name) == 0)
			return domain;
	}

	return NULL;
}

int ect_parse_binary_header(void)
{
	int i, j;
	char *block_name;
	void *address;
	unsigned int length, offset;

	address = (void *)g_address;
	if (address == NULL)
		return -1;

	g_ect_header = kzalloc(sizeof(struct ect_header));

	ect_parse_integer(&address, g_ect_header->sign);
	ect_parse_integer(&address, g_ect_header->version);
	ect_parse_integer(&address, &g_ect_header->total_size);
	ect_parse_integer(&address, &g_ect_header->num_of_header);

	g_base_alloc += g_ect_header->total_size + sizeof(unsigned int);
	align_address();

	if (memcmp_internal(g_ect_header->sign, g_signature, sizeof(g_signature) - 1))
		return -1;

	for (i = 0; i < g_ect_header->num_of_header; ++i) {
		if (ect_parse_string(&address, &block_name, &length)) {
			return -1;
		}

		ect_parse_integer(&address, &offset);

		for (j = 0; j < ARRAY_SIZE(ect_list); ++j) {
			if (strncmp_internal(block_name, ect_list[j].block_name, ect_list[j].block_name_length) != 0)
				continue;

			if (ect_list[j].parser((void *)g_address + offset, ect_list + j))
				return -1;
		}
	}

	return 0;
}

int ect_check_valid_header(void *default_address, void *test_address)
{
	struct ect_header *default_header, *test_header;

	default_header = (struct ect_header *)default_address;
	test_header = (struct ect_header *)test_address;

	if (memcmp_internal(test_header->sign, g_signature, sizeof(g_signature) - 1))
		return -1;

	if (memcmp_internal(test_header->version, default_header->version, sizeof(test_header->version) - 1))
		return -1;

	return 0;
}
