/*
 * Copyright@ Samsung Electronics Co. LTD
 *
 * This software is proprietary of Samsung Electronics.
 * No part of this software, either material or conceptual may be copied or distributed, transmitted,
 * transcribed, stored in a retrieval system or translated into any human or computer language in any form by any means,
 * electronic, mechanical, manual or otherwise, or disclosed
 * to third parties without the express written permission of Samsung Electronics.


 * Alternatively, this program is free software in case of open source project
 * you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.

*/

#include "bl2_common.h"
#include "pmic_S2MPW01.h"
#include "asv.h"
#include "ect_parser.h"
#include "speedy.h"

enum buck1_asv_index {
	CPUCL0,
	G3D,
	MIF,
	INT,
	DISP,
	CAM,
	END_BUCK1_ASV_INDEX
};

int get_boot_level(char *domain_name, int *boot_voltage, int asv_group)
{
	void *dvfs_block, *asv_block, *margin_block;
	struct ect_dvfs_domain *dvfs_domain;
	struct ect_voltage_domain *asv_domain;
	struct ect_voltage_table *asv_table;
	struct ect_margin_domain *margin_domain;
	int boot_level_idx = 0;

	dvfs_block = ect_get_block(BLOCK_DVFS);
	if (dvfs_block == NULL)
		return -1;

	asv_block = ect_get_block(BLOCK_ASV);
	if (asv_block == NULL)
		return -1;

	margin_block = ect_get_block(BLOCK_MARGIN);

	dvfs_domain = ect_dvfs_get_domain(dvfs_block, domain_name);
	if (dvfs_domain == NULL)
		return -1;

	asv_domain = ect_asv_get_domain(asv_block, domain_name);
	if (asv_domain == NULL)
		return -1;

	if (chip_info.nAsv_TABLE >= asv_domain->num_of_table)
		chip_info.nAsv_TABLE = asv_domain->num_of_table - 1;

	margin_domain = ect_margin_get_domain(margin_block, domain_name);

	if (dvfs_domain->boot_level_idx == -1)
		return -1;

	boot_level_idx = dvfs_domain->boot_level_idx;

	asv_table = &asv_domain->table_list[chip_info.nAsv_TABLE];

	if (asv_table->voltages != NULL)
		*boot_voltage = asv_table->voltages[boot_level_idx * asv_domain->num_of_group + asv_group];
	else if (asv_table->voltages_step != NULL)
		*boot_voltage = asv_table->voltages_step[boot_level_idx * asv_domain->num_of_group + asv_group]
				* asv_table->volt_step;
	else
		return -1;


	if (margin_domain != NULL) {
		if (margin_domain->offset != NULL)
			*boot_voltage += margin_domain->offset[boot_level_idx * margin_domain->num_of_group + asv_group];
		else if (margin_domain->offset_compact != NULL)
			*boot_voltage += margin_domain->offset_compact[boot_level_idx * margin_domain->num_of_group + asv_group]
						* margin_domain->volt_step;
	}

	*boot_voltage = (*boot_voltage + 500) / 1000;

	return 0;
}

void pmic_init (void)
{
	int buck1_volt = 0;
	int buck1_volts[END_BUCK1_ASV_INDEX] = {0};
	int buck1_cp_volt = 0;
	int i;
	unsigned char val;

	if (get_boot_level("dvfs_cpucl0", &buck1_volts[CPUCL0], chip_info.nAsv_CPUCL0))
		while(1);

	if (get_boot_level("dvfs_g3d", &buck1_volts[G3D], chip_info.nAsv_G3D))
		while(1);

	if (get_boot_level("dvfs_mif", &buck1_volts[MIF], chip_info.nAsv_MIF))
		while(1);

	if (get_boot_level("dvfs_int", &buck1_volts[INT], chip_info.nAsv_INT))
		while(1);

	if (get_boot_level("dvfs_disp", &buck1_volts[DISP], chip_info.nAsv_CAM_DISP))
		while(1);

	if (get_boot_level("dvfs_cam", &buck1_volts[CAM], chip_info.nAsv_CAM_DISP))
		while(1);

	if (get_boot_level("dvs_cp", &buck1_cp_volt, chip_info.nAsv_CP))
		while(1);

	for(i = 0; i < END_BUCK1_ASV_INDEX; i++) {
		if( buck1_volt < buck1_volts[i])
			buck1_volt = buck1_volts[i];
	}

	speedy_write(S2MPW01_PM_ADDR, S2MPW01_PM_BUCK1_OUT1, WR_BUCK_VOLT(buck1_cp_volt));
	speedy_write(S2MPW01_PM_ADDR, S2MPW01_PM_BUCK1_OUT2, WR_BUCK_VOLT(buck1_volt));
	speedy_write(S2MPW01_PM_ADDR, S2MPW01_PM_BUCK1_OUT3, 0x20);

	/* LDO1 OUT1 0.8V for BUCK1-LDO1 sync off mode */
	speedy_read(S2MPW01_PM_ADDR, S2MPW01_PM_LDO1_CTRL1, &val);
	speedy_write(S2MPW01_PM_ADDR, S2MPW01_PM_LDO1_CTRL1, (val & ~LDOOUT_MASK) | 0x10);
	/* LDO1 OUT2 0.6V for LDO1 DVS */
	speedy_read(S2MPW01_PM_ADDR, S2MPW01_PM_LDO1_CTRL2, &val);
	speedy_write(S2MPW01_PM_ADDR, S2MPW01_PM_LDO1_CTRL2, (val & ~LDOOUT_MASK) | 0x08);
}

void pmic_enable_peric_dev(void)
{
	unsigned char reg;

	pmic_gpio_init();

	speedy_init();

	speedy_read(S2MPW01_PM_ADDR, S2MPW01_PM_TIME_CTRL1, &reg);
	reg |= S2MPW01_WRSTBIEN;
	speedy_write(S2MPW01_PM_ADDR, S2MPW01_PM_TIME_CTRL1, reg);

	speedy_read(S2MPW01_PM_ADDR, S2MPW01_PM_CTRL1, &reg);
	reg &= ~S2MPW01_MRSTBEN;
	speedy_write(S2MPW01_PM_ADDR, S2MPW01_PM_CTRL1, reg);

	speedy_read(S2MPW01_PM_ADDR, S2MPW01_PM_CTRL3, &reg);
	reg |= S2MPW01_WRSTEN;
	speedy_write(S2MPW01_PM_ADDR, S2MPW01_PM_CTRL3, reg);

	speedy_read(S2MPW01_PM_ADDR, S2MPW01_PM_RTC_BUF, &reg);
	reg |= S2MPW01_PERI_32KHZ_EN;
	speedy_write(S2MPW01_PM_ADDR, S2MPW01_PM_RTC_BUF, reg);
}
