/**
 * @file  otp.c
 * @brief OTP functions
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * This software is proprietary of Samsung Electronics.
 * No part of this software, either material or conceptual may be copied
 * or distributed, transmitted, transcribed, stored in a retrieval system
 * or translated into any human or computer language in any form by any means,
 * electronic, mechanical, manual or otherwise, or disclosed to third parties
 * without the express written permission of Samsung Electronics.
 *
 * Alternatively, this program is free software in case of open source project
 * you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <common.h>
#include "otp.h"

#define SMC_AARCH64_PREFIX		0xC2000000

static inline uint64_t exynos_cm_smc(uint64_t *cmd, uint64_t *arg1,
				     uint64_t *arg2, uint64_t *arg3)
{
	register uint64_t reg0 __asm__("x0") = *cmd;
	register uint64_t reg1 __asm__("x1") = *arg1;
	register uint64_t reg2 __asm__("x2") = *arg2;
	register uint64_t reg3 __asm__("x3") = *arg3;

	__asm__ volatile (
		"smc	0\n"
		: "+r"(reg0), "+r"(reg1), "+r"(reg2), "+r"(reg3)
	);

	*cmd = reg0;
	*arg1 = reg1;
	*arg2 = reg2;
	*arg3 = reg3;

	return reg0;
}

uint64_t cm_otp_check_secure_boot_status(uint64_t *secure_boot_status)
{
	uint64_t ret = RV_SUCCESS;
	uint64_t r0 = 0;
	uint64_t r1 = 0;
	uint64_t r2 = 0;
	uint64_t r3 = 0;

	r0 = SMC_AARCH64_PREFIX | SMC_CM_OTP_CONTROL;
	r1 = 0;
	r2 = CMD_R_USE_OEM_KEY;
	r3 = 0;

	ret = exynos_cm_smc(&r0, &r1, &r2, &r3);
	if (ret != RV_SUCCESS) {
		printf("[OTP] Secure Boot check fail with ret:0x%x, r2:0x%x, r3:0x%x\n", \
			(uint32_t)ret, (uint32_t)r2, (uint32_t)r3);
		return ret;
	}

	*secure_boot_status = r2;

	return RV_SUCCESS;
}

static uint64_t cm_otp_write_oem_key(uint8_t *oem_key_ptr, uint32_t oem_key_len)
{
	uint32_t oem_key[OTP_OEM_KEY_SIZE / 4];
	uint64_t r0 = 0;
	uint64_t r1 = 0;
	uint64_t r2 = 0;
	uint64_t r3 = 0;
	uint64_t i = 0;
	uint64_t ret = RV_SUCCESS;

	if (oem_key_ptr == NULL) {
		printf("[OTP] invalid OEM_KEY addresss\n");
		return RV_OTP_PARAM_NULL_POINTER;
	}

	if (oem_key_len != OTP_OEM_KEY_SIZE) {
		printf("[OTP] invalid OEM_KEY length\n");
		return RV_OTP_PARAM_INVALID_KEY_LEN;
	}

	memcpy((uint8_t *)oem_key, oem_key_ptr, OTP_OEM_KEY_SIZE);

	for (i = 0; i < OTP_OEM_KEY_SIZE / 8; i++) {
		r0 = SMC_AARCH64_PREFIX | SMC_CM_OTP_CONTROL;
		r1 = (uint64_t)oem_key[i * 2 + 0];
		r2 = CMD_W_OEM_KEY0_0 + i;
		r3 = (uint64_t)oem_key[i * 2 + 1];

		ret = exynos_cm_smc(&r0, &r1, &r2, &r3);
		if (ret != RV_SUCCESS) {
			printf("[OTP] OEM_KEY program fail with ret:0x%x, r2:0x%x, r3:0x%x\n", \
			(uint32_t)ret, (uint32_t)r2, (uint32_t)r3);
			return ret;
		}
	}
	printf("[OTP] OEM_KEY program success\n");
	return ret;
}

uint64_t cm_otp_enable_secure_boot(uint8_t *oem_key_ptr, uint32_t oem_key_len)
{
	uint64_t secure_boot_status = 0;
	uint64_t ret = RV_SUCCESS;

	ret = cm_otp_check_secure_boot_status(&secure_boot_status);
	if (ret != RV_SUCCESS) {
		if (ret == RV_OTP_LDFW_NOT_LOADED)
			printf("[OTP] LDFW has not loaded\n");
		return ret;
	}

	if (secure_boot_status == 1) {
		printf("[OTP] Secure Boot enabled already\n");
		return RV_SUCCESS;
	}

	ret = cm_otp_write_oem_key(oem_key_ptr, oem_key_len);
	if (ret == RV_OTP_OEM_KEY_CMP_FAIL) {
		printf("[OTP] Secure Boot enable: try again\n");
		ret = cm_otp_write_oem_key(oem_key_ptr, oem_key_len);
	}

	if (ret != RV_SUCCESS)
		printf("[OTP] Secure Boot enable: fail: 0x%x\n", (uint32_t)ret);
	else
		printf("[OTP] Secure Boot enable: success\n");

	return ret;
}

uint64_t cm_otp_check_secure_jtag_status(uint64_t *secure_jtag_status)
{
	uint64_t ret = RV_SUCCESS;
	uint64_t r0 = 0;
	uint64_t r1 = 0;
	uint64_t r2 = 0;
	uint64_t r3 = 0;

	r0 = SMC_AARCH64_PREFIX | SMC_CM_OTP_CONTROL;
	r1 = 0;
	r2 = CMD_R_JTAG_LOCK;
	r3 = 0;

	ret = exynos_cm_smc(&r0, &r1, &r2, &r3);
	if (ret != RV_SUCCESS) {
		printf("[OTP] Secure JTAG check fail with ret:0x%x, r2:0x%x, r3:0x%x\n", \
			(uint32_t)ret, (uint32_t)r2, (uint32_t)r3);
		return ret;
	}

	*secure_jtag_status = r2;

	return RV_SUCCESS;
}

static uint64_t cm_otp_write_jtag_key(uint8_t *jtag_key_ptr, uint32_t jtag_key_len)
{
	uint32_t jtag_key[OTP_JTAG_KEY_SIZE / 4];
	uint64_t ret = RV_SUCCESS;
	uint64_t r0 = 0;
	uint64_t r1 = 0;
	uint64_t r2 = 0;
	uint64_t r3 = 0;
	uint64_t i = 0;

	if (jtag_key_ptr == NULL) {
		printf("[OTP] invalid JTAG_KEY addresss\n");
		return RV_OTP_PARAM_NULL_POINTER;
	}

	if (jtag_key_len != OTP_JTAG_KEY_SIZE) {
		printf("[OTP] invalid JTAG_KEY length\n");
		return RV_OTP_PARAM_INVALID_KEY_LEN;
	}

	memcpy((uint8_t *)jtag_key, jtag_key_ptr, OTP_JTAG_KEY_SIZE);

	for (i = 0; i < OTP_JTAG_KEY_SIZE / 8; i++) {
		r0 = SMC_AARCH64_PREFIX | SMC_CM_OTP_CONTROL;
		r1 = (uint64_t)jtag_key[i * 2 + 0];
		r2 = CMD_W_JTAG_KEY_0 + i;
		r3 = (uint64_t)jtag_key[i * 2 + 1];

		ret = exynos_cm_smc(&r0, &r1, &r2, &r3);
		if (ret != RV_SUCCESS) {
			printf("[OTP] JTAG_KEY program fail with ret:0x%x, r2:0x%x, r3:0x%x\n", \
			(uint32_t)ret, (uint32_t)r2, (uint32_t)r3);
			return ret;
		}
	}
	printf("[OTP] JTAG_KEY program success\n");
	return ret;
}

uint64_t cm_otp_enable_secure_jtag(uint8_t *jtag_key_ptr, uint32_t jtag_key_len)
{
	uint64_t ret = RV_SUCCESS;
	uint64_t secure_jtag_status = 0;

	ret = cm_otp_check_secure_jtag_status(&secure_jtag_status);
	if (ret != RV_SUCCESS) {
		if (ret == RV_OTP_LDFW_NOT_LOADED)
			printf("[OTP] LDFW has not loaded\n");
		return ret;
	}

	if (secure_jtag_status == 1) {
		printf("[OTP] Secure JTAG enabled already\n");
		return RV_SUCCESS;
	}

	ret = cm_otp_write_jtag_key(jtag_key_ptr, jtag_key_len);
	if (ret == RV_OTP_JTAG_KEY_CMP_FAIL) {
		printf("[OTP] Secure JTAG enable: try again\n");
		ret = cm_otp_write_jtag_key(jtag_key_ptr, jtag_key_len);
	}

	if (ret != RV_SUCCESS)
		printf("[OTP] Secure JTAG enable: fail: 0x%x\n", (uint32_t)ret);
	else
		printf("[OTP] Secure JTAG enable: success\n");

	return ret;
}
